# micropython ESP32
# SD card image writer

# AUTHOR=EMARD
# LICENSE=BSD

from time import ticks_ms
from machine import SPI, Pin, SDCard
from micropython import const

class sdraw:
  #def __init__(self):
    #print("SD RAW writer")
    #self.init_pinout_sd()

  def stopwatch_start(self):
    self.stopwatch_ms = ticks_ms()
  
  def stopwatch_stop(self, bytes_uploaded):
    elapsed_ms = ticks_ms() - self.stopwatch_ms
    transfer_rate_MBps = 0
    if elapsed_ms > 0:
      transfer_rate_kBps = bytes_uploaded // elapsed_ms
    print("%d bytes uploaded in %d ms (%d kB/s)" % (bytes_uploaded, elapsed_ms, transfer_rate_kBps))

  def open_file(self, filename, gz=False):
    filedata = open(filename, "rb")
    if gz:
      import uzlib
      return uzlib.DecompIO(filedata,31)
    return filedata

  def open_web(self, url, gz=False):
    import socket
    _, _, host, path = url.split('/', 3)
    port = 80
    if ( len(host.split(':')) == 2 ):
      host, port = host.split(':', 2)
    print("host = ",host," port = ", port, " path = ", path)
    addr = socket.getaddrinfo(host, port)[0][-1]
    s = socket.socket()
    s.connect(addr)
    s.send(bytes('GET /%s HTTP/1.0\r\nHost: %s\r\nAccept:  image/*\r\n\r\n' % (path, host), 'utf8'))
    for i in range(100): # read first 100 lines searching for
      if len(s.readline()) < 3: # first empty line (contains "\r\n")
        break
    if gz:
      import uzlib
      return uzlib.DecompIO(s,31)
    return s

  def sd_open(self):
    self.sd = SDCard(slot=3)

  def sd_close(self):
    self.sd.deinit()
    for i in bytearray([2,4,12,13,14,15]):
      p=Pin(i,Pin.IN)
      a=p.value()
      del p, a
    del self.sd

  def sd_check_param(self, addr):
    if addr & 0x1FF:
      print("parameter must be rounded to block_size = 512 bytes")
      return False
    return True

  # negative addr means reference from end of the card  
  def sd_wrapaddr(self, addr):
    if addr >= 0:
      return addr
    cardsize = self.sd.ioctl(4,0)*0x200
    return cardsize+addr

  def sd_read(self, data, addr=0):
    if not self.sd_check_param(addr) or not self.sd_check_param(len(data)):
      return False
    self.sd_open()
    self.sd.readblocks(self.sd_wrapaddr(addr)//0x200,data)
    self.sd_close()
    return True

  def sd_write_stream(self, filedata, addr=0, blocksize=16384):
    if not self.sd_check_param(addr):
      return False
    bytes_uploaded = 0
    self.sd_open()
    addr=self.sd_wrapaddr(addr)
    nearend=self.sd_wrapaddr(-blocksize)
    self.stopwatch_start()
    block = bytearray(blocksize)
    while True:
      waddr=addr+bytes_uploaded
      if waddr >= nearend and len(block) > 0x200:
        block = bytearray(0x200)
      if filedata.readinto(block):
        self.sd.writeblocks(waddr//0x200,block)
        bytes_uploaded += len(block)
      else:
        break
    self.stopwatch_stop(bytes_uploaded)
    self.sd_close()
    return True

def read(addr=0, length=512):
  data = bytearray(length)
  if sdraw().sd_read(data, addr):
    return data
  else:
    return False

def write(filepath, addr=0):
  gz=filepath.endswith(".gz")
  if filepath.startswith("http://") or filepath.startswith("/http:/"):
    filedata = sdraw().open_web(filepath, gz)
  else:
    filedata = sdraw().open_file(filepath, gz)
  if filedata:
    if gz:
      return sdraw().sd_write_stream(filedata,addr,blocksize=4096)
    else:
      return sdraw().sd_write_stream(filedata,addr,blocksize=16384)
  return False

def help():
  print("usage:")
  print("sdraw.write(\"http://192.168.4.2/sdcard.img\", addr=0) # to start of SD")
  print("sdraw.read(addr=0, length=512) # from start of SD")
  print("sdraw.read(-1024) # from 1024 bytes before end of SD")



import network
import time

def read_profiles():
    with open("wifiman.conf") as f:
        lines = f.readlines()
    profiles = {}
    for line in lines:
        ssid, password = line.strip("\n").split(":")
        profiles[ssid] = password
    return profiles

wlan_sta = network.WLAN(network.STA_IF)

def get_connection():
    """return a working WLAN(STA_IF) instance or None"""

    # First check if there already is any connection:
    if wlan_sta.isconnected():
        return wlan_sta

    connected = False
    try:
        # ESP connecting to WiFi takes time, wait a bit and try again:
        time.sleep(3)
        if wlan_sta.isconnected():
            return wlan_sta

        # Read known network profiles from file
        profiles = read_profiles()

        # Search WiFis in range
        wlan_sta.active(True)
        networks = wlan_sta.scan()

        AUTHMODE = {0: "open", 1: "WEP", 2: "WPA-PSK", 3: "WPA2-PSK", 4: "WPA/WPA2-PSK"}
        for ssid, bssid, channel, rssi, authmode, hidden in sorted(networks, key=lambda x: x[3], reverse=True):
            ssid = ssid.decode('utf-8')
            encrypted = authmode > 0
            print("ssid: %s chan: %d rssi: %d authmode: %s" % (ssid, channel, rssi, AUTHMODE.get(authmode, '?')))
            if encrypted:
                if ssid in profiles:
                    password = profiles[ssid]
                    connected = do_connect(ssid, password)
            else:  # open
                connected = do_connect(ssid, None)
            if connected:
                break

    except OSError as e:
        print("exception", str(e))

    return wlan_sta

def do_connect(ssid, password):
    wlan_sta.active(True)
    if wlan_sta.isconnected():
        return None
    wlan_sta.connect(ssid, password)
    for retry in range(100):
        connected = wlan_sta.isconnected()
        if connected:
            break
        time.sleep(0.1)
    return connected

get_connection()



#
# Small ftp server for ESP8266 Micropython
# Based on the work of chrisgp - Christopher Popp and pfalcon - Paul Sokolovsky
#
# The server accepts passive mode only. It runs in background.
# Start the server with:
#
# import uftpd
# uftpd.start([port = 21][, verbose = level])
#
# port is the port number (default 21)
# verbose controls the level of printed activity messages, values 0, 1, 2
#
# Copyright (c) 2016 Christopher Popp (initial ftp server framework)
# Copyright (c) 2016 Paul Sokolovsky (background execution control structure)
# Copyright (c) 2016 Robert Hammelrath (putting the pieces together and a
# few extensions)
# Distributed under MIT License
#
import socket, network, os, io
from gc import collect
from time import sleep_ms, localtime
from micropython import alloc_emergency_exception_buf
from machine import SDCard, Pin

# constant definitions
_CHUNK_SIZE = const(1024)
_SO_REGISTER_HANDLER = const(20)
_COMMAND_TIMEOUT = const(300)
_DATA_TIMEOUT = const(100)
_DATA_PORT = const(13333)

# Global variables
ftpsocket = None
datasocket = None
client_list = []
verbose_l = 0
client_busy = False
# Interfaces: (IP-Address (string), IP-Address (integer), Netmask (integer))
AP_addr = ("0.0.0.0", 0, 0xffffff00)
STA_addr = ("0.0.0.0", 0, 0xffffff00)

_month_name = ("", "Jan", "Feb", "Mar", "Apr", "May", "Jun",
               "Jul", "Aug", "Sep", "Oct", "Nov", "Dec")

def mount():
  global SD
  try:
    SD = SDCard(slot=3)
    os.mount(SD,"/sd")
    return True
  except:
    return False

def umount():
  global SD
  try:
    os.umount("/sd")
    try:
      SD.deinit()
      del SD
    except:
      pass
    # let all SD pins be inputs
    for i in bytearray([2,4,12,13,14,15]):
      p = Pin(i,Pin.IN)
      a = p.value()
      del p, a
    return True
  except:
    return False

def passthru():
  import ecp5
  ecp5.passthru()

# to capture output of exec()
class DUP(io.IOBase):
    def __init__(self, s):
        self.s = s

    def write(self, data):
        self.s += data
        return len(data)

    def readinto(self, data):
        return 0

class FTP_client:
  def __init__(self, ftpsocket):
    global AP_addr, STA_addr
    self.command_client, self.remote_addr = ftpsocket.accept()
    self.remote_addr = self.remote_addr[0]
    self.command_client.settimeout(_COMMAND_TIMEOUT)
    log_msg(1, "FTP Command connection from:", self.remote_addr)
    self.command_client.setsockopt(socket.SOL_SOCKET,
                                   _SO_REGISTER_HANDLER,
                                   self.exec_ftp_command)
    self.command_client.sendall("220 Hello, this is the ULX3S.\r\n")
    self.cwd = '/'
    self.fromname = None
    # self.logged_in = False
    self.act_data_addr = self.remote_addr
    self.DATA_PORT = 20
    self.active = True
    # check which interface was used by comparing the caller's ip
    # adress with the ip adresses of STA and AP; consider netmask;
    # select IP address for passive mode
    if ((AP_addr[1] & AP_addr[2]) ==
       (num_ip(self.remote_addr) & AP_addr[2])):
        self.pasv_data_addr = AP_addr[0]
    elif ((STA_addr[1] & STA_addr[2]) ==
          (num_ip(self.remote_addr) & STA_addr[2])):
        self.pasv_data_addr = STA_addr[0]
    elif ((AP_addr[1] == 0) and (STA_addr[1] != 0)):
        self.pasv_data_addr = STA_addr[0]
    elif ((AP_addr[1] != 0) and (STA_addr[1] == 0)):
        self.pasv_data_addr = AP_addr[0]
    else:
        self.pasv_data_addr = "0.0.0.0"  # Invalid value

  def send_list_data(self, path, data_client, full):
    try:
      for fname in os.listdir(path):
        data_client.sendall(self.make_description(path, fname, full))
    except:  # path may be a file name or pattern
      path, pattern = self.split_path(path)
      try:
        for fname in os.listdir(path):
          if self.fncmp(fname, pattern):
            data_client.sendall(
              self.make_description(path, fname, full))
      except:
          pass

  def make_description(self, path, fname, full):
    global _month_name
    if full:
      stat = os.stat(self.get_absolute_path(path, fname))
      file_permissions = ("drwxr-xr-x"
                          if (stat[0] & 0o170000 == 0o040000)
                          else "-rw-r--r--")
      file_size = stat[6]
      tm = localtime(stat[7] if stat[7]<0x80000000 else stat[7]-0x100000000)
      if tm[0] != localtime()[0]:
        description = "{} 1 owner group {:>10} {} {:2} {:>5} {}\r\n".\
          format(file_permissions, file_size,
                 _month_name[tm[1]], tm[2], tm[0], fname)
      else:
        description = "{} 1 owner group {:>10} {} {:2} {:02}:{:02} {}\r\n".\
          format(file_permissions, file_size,
                 _month_name[tm[1]], tm[2], tm[3], tm[4], fname)
    else:
      description = fname + "\r\n"
    return description

  def send_file_data(self, path, data_client):
    with open(path,"rb") as file:
      chunk = file.read(_CHUNK_SIZE)
      while len(chunk) > 0:
        data_client.sendall(chunk)
        chunk = file.read(_CHUNK_SIZE)
      data_client.close()

  def save_file_data(self, path, data_client, mode):
    with open(path, mode) as file:
      chunk = data_client.recv(_CHUNK_SIZE)
      while len(chunk) > 0:
        file.write(chunk)
        chunk = data_client.recv(_CHUNK_SIZE)
      data_client.close()

  def get_absolute_path(self, cwd, payload):
    # Just a few special cases "..", "." and ""
    # If payload start's with /, set cwd to /
    # and consider the remainder a relative path
    if payload.startswith('/'):
      cwd = "/"
    for token in payload.split("/"):
      if token == '..':
        cwd = self.split_path(cwd)[0]
      elif token != '.' and token != '':
        if cwd == '/':
          cwd += token
        else:
          cwd = cwd + '/' + token
    return cwd

  def split_path(self, path):  # instead of path.rpartition('/')
    tail = path.split('/')[-1]
    head = path[:-(len(tail) + 1)]
    return ('/' if head == '' else head, tail)

  # compare fname against pattern. Pattern may contain
  # the wildcards ? and *.
  def fncmp(self, fname, pattern):
    pi = 0
    si = 0
    while pi < len(pattern) and si < len(fname):
      if (fname[si] == pattern[pi]) or (pattern[pi] == '?'):
        si += 1
        pi += 1
      else:
        if pattern[pi] == '*':  # recurse
          if pi == len(pattern.rstrip("*?")):  # only wildcards left
            return True
          while si < len(fname):
            if self.fncmp(fname[si:], pattern[pi + 1:]):
              return True
            else:
              si += 1
          return False
        else:
          return False
    if pi == len(pattern.rstrip("*")) and si == len(fname):
      return True
    else:
      return False

  def open_dataclient(self):
    if self.active:  # active mode
      data_client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
      data_client.settimeout(_DATA_TIMEOUT)
      data_client.connect((self.act_data_addr, self.DATA_PORT))
      log_msg(1, "FTP Data connection with:", self.act_data_addr)
    else:  # passive mode
      data_client, data_addr = datasocket.accept()
      log_msg(1, "FTP Data connection with:", data_addr[0])
    return data_client

  def exec_ftp_command(self, cl):
    global datasocket
    global client_busy
    global my_ip_addr

    try:
      collect()

      data = cl.readline().decode("utf-8").rstrip("\r\n")

      if len(data) <= 0:
          # No data, close
          # This part is NOT CLEAN; there is still a chance that a
          # closing data connection will be signalled as closing
          # command connection
          log_msg(1, "*** No data, assume QUIT")
          close_client(cl)
          return

      if client_busy:  # check if another client is busy
          cl.sendall("400 Device busy.\r\n")  # tell so the remote client
          return  # and quit
      client_busy = True  # now it's my turn

      # check for log-in state may done here, like
      # if self.logged_in == False and not command in\
      #    ("USER", "PASS", "QUIT"):
      #    cl.sendall("530 Not logged in.\r\n")
      #    return

      command = data.split()[0].upper()
      payload = data[len(command):].lstrip()  # partition is missing
      path = self.get_absolute_path(self.cwd, payload)
      log_msg(1, "Command={}, Payload={}".format(command, payload))

      if command == "USER":
        # self.logged_in = True
        cl.sendall("230 Logged in.\r\n")
        # If you want to see a password,return
        #   "331 Need password.\r\n" instead
        # If you want to reject an user, return
        #   "530 Not logged in.\r\n"
      elif command == "PASS":
        # you may check here for a valid password and return
        # "530 Not logged in.\r\n" in case it's wrong
        # self.logged_in = True
        cl.sendall("230 Logged in.\r\n")
      elif command == "SYST":
        cl.sendall("215 UNIX Type: L8\r\n")
      elif command in ("TYPE", "NOOP", "ABOR"):  # just accept & ignore
        cl.sendall('200 OK\r\n')
      elif command == "QUIT":
        cl.sendall('221 Bye.\r\n')
        close_client(cl)
      elif command == "PWD" or command == "XPWD":
        cl.sendall('257 "{}"\r\n'.format(self.cwd))
      elif command == "CWD" or command == "XCWD":
        try:
          if (os.stat(path)[0] & 0o170000) == 0o040000:
            self.cwd = path
            cl.sendall('250 OK\r\n')
          else:
            cl.sendall('550 Fail\r\n')
        except:
          cl.sendall('550 Fail\r\n')
      elif command == "PASV":
        cl.sendall('227 Entering Passive Mode ({},{},{}).\r\n'.format(
          self.pasv_data_addr.replace('.', ','),
          _DATA_PORT >> 8, _DATA_PORT % 256))
        self.active = False
      elif command == "PORT":
        items = payload.split(",")
        if len(items) >= 6:
          self.act_data_addr = '.'.join(items[:4])
          if self.act_data_addr == "127.0.1.1":
            # replace by command session addr
            self.act_data_addr = self.remote_addr
          self.DATA_PORT = int(items[4]) * 256 + int(items[5])
          cl.sendall('200 OK\r\n')
          self.active = True
        else:
            cl.sendall('504 Fail\r\n')
      elif command == "LIST" or command == "NLST":
        if payload.startswith("-"):
          option = payload.split()[0].lower()
          path = self.get_absolute_path(
                 self.cwd, payload[len(option):].lstrip())
        else:
          option = ""
        try:
          data_client = self.open_dataclient()
          cl.sendall("150 Directory listing:\r\n")
          self.send_list_data(path, data_client,
            command == "LIST" or 'l' in option)
          cl.sendall("226 Done.\r\n")
          data_client.close()
        except:
          cl.sendall('550 Fail\r\n')
          if data_client is not None:
            data_client.close()
      elif command == "RETR":
        try:
          data_client = self.open_dataclient()
          cl.sendall("150 Opened data connection.\r\n")
          self.send_file_data(path, data_client)
          # if the next statement is reached,
          # the data_client was closed.
          data_client = None
          cl.sendall("226 Done.\r\n")
        except:
          cl.sendall('550 Fail\r\n')
          if data_client is not None:
            data_client.close()
      elif command == "STOR" or command == "APPE":
        result = False
        try:
          data_client = self.open_dataclient()
          cl.sendall("150 Opened data connection.\r\n")
          if path == "/fpga":
            import ecp5
            ecp5.prog_stream(data_client,_CHUNK_SIZE)
            result = ecp5.prog_close()
            data_client.close()
          elif path.startswith("/flash@"):
            import ecp5
            dummy, addr = path.split("@")
            addr = int(addr)
            result = ecp5.flash_stream(data_client,addr)
            ecp5.flash_close()
            del addr, dummy
            data_client.close()
          elif path.startswith("/sd@"):
            import sdraw
            dummy, addr = path.split("@")
            addr = int(addr)
            sd_raw = sdraw.sdraw()
            result = sd_raw.sd_write_stream(data_client,addr)
            del sd_raw, addr, dummy
            data_client.close()
          else:
            self.save_file_data(path, data_client,
                                "w" if command == "STOR" else "a")
            result = True
          # if the next statement is reached,
          # the data_client was closed.
          data_client = None
        except:
          if data_client is not None:
            data_client.close()
        if result:
          cl.sendall("226 Done.\r\n")
        else:
          cl.sendall('550 Fail\r\n')
        del result
      elif command == "SIZE":
        try:
          cl.sendall('213 {}\r\n'.format(os.stat(path)[6]))
        except:
          cl.sendall('550 Fail\r\n')
      elif command == "STAT":
        if payload == "":
          cl.sendall("211-Connected to ({})\r\n"
                     "    Data address ({})\r\n"
                     "    TYPE: Binary STRU: File MODE: Stream\r\n"
                     "    Session timeout {}\r\n"
                     "211 Client count is {}\r\n".format(
                      self.remote_addr, self.pasv_data_addr,
                      _COMMAND_TIMEOUT, len(client_list)))
        else:
          cl.sendall("213-Directory listing:\r\n")
          self.send_list_data(path, cl, True)
          cl.sendall("213 Done.\r\n")
      elif command == "DELE":
        try:
          os.remove(path)
          cl.sendall('250 OK\r\n')
        except:
          cl.sendall('550 Fail\r\n')
      elif command == "RNFR":
        try:
          # just test if the name exists, exception if not
          os.stat(path)
          self.fromname = path
          cl.sendall("350 Rename from\r\n")
        except:
          cl.sendall('550 Fail\r\n')
      elif command == "RNTO":
        try:
          os.rename(self.fromname, path)
          cl.sendall('250 OK\r\n')
        except:
          cl.sendall('550 Fail\r\n')
        self.fromname = None
      elif command == "CDUP" or command == "XCUP":
        self.cwd = self.get_absolute_path(self.cwd, "..")
        cl.sendall('250 OK\r\n')
      elif command == "RMD" or command == "XRMD":
        try:
          os.rmdir(path)
          cl.sendall('250 OK\r\n')
        except:
          cl.sendall('550 Fail\r\n')
      elif command == "MKD" or command == "XMKD":
        try:
          os.mkdir(path)
          cl.sendall('250 OK\r\n')
        except:
          cl.sendall('550 Fail\r\n')
      elif command == "SITE":
        if path.endswith(".bit") or path.endswith(".bit.gz"):
          try:
            import ecp5
            if ecp5.prog(path, close=False):
              if path.startswith("/sd/"):
                try:
                  umount()
                  cl.sendall('250-umount /sd OK\r\n')
                except:
                  cl.sendall('550-umount /sd Fail\r\n')
              if ecp5.prog_close():
                cl.sendall('250 OK\r\n')
              else:
                cl.sendall('550 Fail\r\n')
            else:
              cl.sendall('550 Fail\r\n')
          except:
            cl.sendall('550 Fail\r\n')
        else:
          try:
            stdout=bytearray()
            os.dupterm(DUP(stdout))
            exec(payload)
            os.dupterm(None)
            cl.sendall('250-\r\n')
            for line in stdout.decode("utf-8").splitlines():
              cl.sendall(line+'\r\n')
            cl.sendall('250 OK\r\n')
            del stdout
          except:
            cl.sendall('550 Fail '+payload+'\r\n')
      else:
        cl.sendall("502 Unsupported command.\r\n")
        # log_msg(2,
        #  "Unsupported command {} with payload {}".format(command,
        #  payload))
    # handle unexpected errors
    except Exception as err:
      log_msg(1, "Exception in exec_ftp_command: {}".format(err))
    # tidy up before leaving
    client_busy = False

def log_msg(level, *args):
  global verbose_l
  if verbose_l >= level:
    print(*args)

# close client and remove it from the list
def close_client(cl):
  cl.setsockopt(socket.SOL_SOCKET, _SO_REGISTER_HANDLER, None)
  cl.close()
  for i, client in enumerate(client_list):
    if client.command_client == cl:
      del client_list[i]
      break

def accept_ftp_connect(ftpsocket):
  # Accept new calls for the server
  try:
    client_list.append(FTP_client(ftpsocket))
  except:
    log_msg(1, "Attempt to connect failed")
    # try at least to reject
    try:
      temp_client, temp_addr = ftpsocket.accept()
      temp_client.close()
    except:
      pass

def num_ip(ip):
  items = ip.split(".")
  return (int(items[0]) << 24 | int(items[1]) << 16 |
          int(items[2]) <<  8 | int(items[3]))

def stop():
  global ftpsocket, datasocket
  global client_list
  global client_busy

  for client in client_list:
    client.command_client.setsockopt(socket.SOL_SOCKET, _SO_REGISTER_HANDLER, None)
    client.command_client.close()
  del client_list
  client_list = []
  client_busy = False
  if ftpsocket is not None:
    ftpsocket.setsockopt(socket.SOL_SOCKET, _SO_REGISTER_HANDLER, None)
    ftpsocket.close()
  if datasocket is not None:
    datasocket.close()

# start listening for ftp connections on port 21
def start(port=21, verbose=0, splash=True):
  global ftpsocket, datasocket
  global verbose_l
  global client_list
  global client_busy
  global AP_addr, STA_addr

  alloc_emergency_exception_buf(100)
  verbose_l = verbose
  client_list = []
  client_busy = False

  ftpsocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  datasocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

  ftpsocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
  datasocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

  ftpsocket.bind(('0.0.0.0', port))
  datasocket.bind(('0.0.0.0', _DATA_PORT))

  ftpsocket.listen(0)
  datasocket.listen(0)

  datasocket.settimeout(10)
  ftpsocket.setsockopt(socket.SOL_SOCKET, _SO_REGISTER_HANDLER, accept_ftp_connect)

  wlan = network.WLAN(network.AP_IF)
  if wlan.active():
    ifconfig = wlan.ifconfig()
    # save IP address string and numerical values of IP adress and netmask
    AP_addr = (ifconfig[0], num_ip(ifconfig[0]), num_ip(ifconfig[1]))
    if splash:
      print("FTP server started on {}:{}".format(ifconfig[0], port))
  wlan = network.WLAN(network.STA_IF)
  if wlan.active():
    ifconfig = wlan.ifconfig()
    # save IP address string and numerical values of IP adress and netmask
    STA_addr = (ifconfig[0], num_ip(ifconfig[0]), num_ip(ifconfig[1]))
    if splash:
      print("FTP server started on {}:{}".format(ifconfig[0], port))

def restart(port=21, verbose=0, splash=True):
  stop()
  sleep_ms(200)
  start(port, verbose, splash)

start(splash=True)
collect()



import os

def printfile(name):
  try:
    f=open(name,"r")
    user_pass=f.read()
    print(user_pass)
    f.close()
  except:
    print("file '%s' not found." % name)

printfile("wifiman.conf")
yn=input("delete old and create new WiFi users:passwords (n/y)? ")
if yn.startswith("y"):
  f=open("wifiman.conf","w")
  f.close()
  print("all WiFi users:passwords deleted")
yn="y"
while yn.startswith("y"):
  yn=input("add WiFi user:password (n/y)? ")
  if yn.startswith("y"):
    user_pass=input("enter WiFi user:password> ")
    if user_pass.find(":")>0:
      f=open("wifiman.conf","a")
      f.write(user_pass)
      f.write("\n")
      f.close()
  printfile("wifiman.conf")

def main():
  f=open("main.py","w")
  f.write(
"try:\n"
"  import wifiman\n"
"except:\n"
"  print('no WiFi')\n"
"import gc\n"
"gc.collect()\n"
"import uftpd\n"
"from ntptime import settime\n"
"try:\n"
"  settime()\n"
"except:\n"
"  print('NTP not available')\n"
  )
  f.close()

yn=input("overwrite 'main.py' to run FTP server at boot (n/y)? ")
if yn.startswith("y"):
  main()
  print("main.py overwritten")

def pins_v20():
  f=open("jtagpin.py","w")
  f.write(
"# ULX3S v3.0.x or v2.x.x\n"
"tms=const(21)\n"
"tck=const(18)\n"
"tdi=const(23)\n"
"tdo=const(19)\n"
"tcknc=const(17)\n"
"led=const(5)\n"
  )
  f.close()

def pins_v31():
  f=open("jtagpin.py","w")
  f.write(
"# ULX3S v3.1.x\n"
"tms=const(5)\n"
"tck=const(18)\n"
"tdi=const(23)\n"
"tdo=const(34)\n"
"tcknc=const(21)\n"
"led=const(19)\n"
  )
  f.close()

def pins_fjc():
  f=open("jtagpin.py","w")
  f.write(
"# FJC-ESP32-V0r2\n"
"tms=const(4)\n"
"tck=const(16)\n"
"tdi=const(15)\n"
"tdo=const(2)\n"
"tcknc=const(21)\n"
"led=const(19)\n"
  )
  f.close()

printfile("jtagpin.py")
yn=input("change JTAG pinout (n/y)? ")
if yn.startswith("y"):
  print("0: ULX3S v3.0.x or v2.x.x")
  print("1: ULX3S v3.1.x")
  print("2: FJC-ESP32-V0r2")
  pinout=input("select pinout (0-2)> ")
  if pinout.startswith("0"):
    pins_v20()
  if pinout.startswith("1"):
    pins_v31()
  if pinout.startswith("2"):
    pins_fjc()
  printfile("jtagpin.py")



# micropython ESP32
# ECP5 JTAG FLASH protection tool

# AUTHOR=EMARD
# LICENSE=BSD

from time import ticks_ms, sleep_ms
from machine import SPI, Pin
from micropython import const
from struct import unpack
from uctypes import addressof
import jtagpin
#from gc import collect

#@micropython.viper
#def init_reverse_bits():
#  #p8rb=ptr8(addressof(rb))
#  p8rb=memoryview(rb)
#  for i in range(256):
#    v=i
#    r=0
#    for j in range(8):
#      r<<=1
#      r|=v&1
#      v>>=1
#    p8rb[i]=r

spi_freq = const(20000000) # Hz JTAG clk frequency
# -1 for JTAG over SOFT SPI slow, compatibility
#  1 or 2 for JTAG over HARD SPI fast
#  2 is preferred as it has default pinout wired
spi_channel = const(2) # -1 soft, 1:sd, 2:jtag
#flash_read_size = const(2048)
#flash_write_size = const(256)
#flash_erase_size = const(4096)
#flash_erase_cmd = { 4096:0x20, 32768:0x52, 65536:0xD8, 262144:0xD8 } # erase commands from FLASH PDF
#flash_era = bytearray([flash_erase_cmd[flash_erase_size],0,0])
#flash_req=bytearray(4)
read_status=bytearray([5])
status=bytearray(1)
#rb=bytearray(256) # reverse bits
#init_reverse_bits()
discard=const(0) # discard running bitstream

def bitbang_jtag_on():
  global tck,tms,tdi,tdo,led
  led=Pin(jtagpin.led,Pin.OUT)
  tms=Pin(jtagpin.tms,Pin.OUT)
  tck=Pin(jtagpin.tck,Pin.OUT)
  tdi=Pin(jtagpin.tdi,Pin.OUT)
  tdo=Pin(jtagpin.tdo,Pin.IN)

def bitbang_jtag_off():
  global tck,tms,tdi,tdo,led
  led=Pin(jtagpin.led,Pin.IN)
  tms=Pin(jtagpin.tms,Pin.IN)
  tck=Pin(jtagpin.tck,Pin.IN)
  tdi=Pin(jtagpin.tdi,Pin.IN)
  tdo=Pin(jtagpin.tdo,Pin.IN)
  a = led.value()
  a = tms.value()
  a = tck.value()
  a = tdo.value()
  a = tdi.value()
  del led
  del tms
  del tck
  del tdi
  del tdo

# initialize both hardware accelerated SPI
# software SPI on the same pins
def spi_jtag_on():
  global hwspi,swspi
  hwspi=SPI(spi_channel, baudrate=spi_freq, polarity=1, phase=1, bits=8, firstbit=SPI.MSB, sck=Pin(jtagpin.tck), mosi=Pin(jtagpin.tdi), miso=Pin(jtagpin.tdo))
  swspi=SPI(-1, baudrate=spi_freq, polarity=1, phase=1, bits=8, firstbit=SPI.MSB, sck=Pin(jtagpin.tck), mosi=Pin(jtagpin.tdi), miso=Pin(jtagpin.tdo))

def spi_jtag_off():
  global hwspi,swspi
  hwspi.deinit()
  del hwspi
  swspi.deinit()
  del swspi

# print bytes reverse - appears the same as in SVF file
#def print_hex_reverse(block, head="", tail="\n"):
#  print(head, end="")
#  for n in range(len(block)):
#    print("%02X" % block[len(block)-n-1], end="")
#  print(tail, end="")

@micropython.viper
def send_tms(val:int):
  if val:
    tms.on()
  else:
    tms.off()
  tck.off()
  tck.on()

# exit 1 DR -> select DR scan
def send_tms0111():
  send_tms(0) # -> pause DR
  send_tms(1) # -> exit 2 DR
  send_tms(1) # -> update DR
  send_tms(1) # -> select DR scan

@micropython.viper
def send_read_buf_lsb1st(buf, last:int, w:ptr8):
  p = ptr8(addressof(buf))
  l = int(len(buf))
  val = 0
  tms.off()
  for i in range(l-1):
    byte = 0
    val = p[i]
    for nf in range(8):
      if (val >> nf) & 1:
        tdi.on()
      else:
        tdi.off()
      tck.off()
      tck.on()
      if tdo.value():
        byte |= 1 << nf
    if int(w):
      w[i] = byte # write byte
  byte = 0
  val = p[l-1] # read last byte
  for nf in range(7): # first 7 bits
    if (val >> nf) & 1:
      tdi.on()
    else:
      tdi.off()
    tck.off()
    tck.on()
    if tdo.value():
      byte |= 1 << nf
  # last bit
  if last:
    tms.on()
  if (val >> 7) & 1:
    tdi.on()
  else:
    tdi.off()
  tck.off()
  tck.on()
  if tdo.value():
    byte |= 1 << 7
  if int(w):
    w[l-1] = byte # write last byte

@micropython.viper
def send_int_msb1st(val:int, last:int, bits:int):
  tms.off()
  for nf in range(bits-1):
    if (val >> (7-nf)) & 1:
      tdi.on()
    else:
      tdi.off()
    tck.off()
    tck.on()
  if last:
    tms.on()
  if val & 1:
    tdi.on()
  else:
    tdi.off()
  tck.off()
  tck.on()

# TAP to "reset" state
@micropython.viper
def reset_tap():
  for n in range(6):
    send_tms(1) # -> Test Logic Reset

# TAP should be in "idle" state
# TAP returns to "select DR scan" state
@micropython.viper
def runtest_idle(count:int, duration_ms:int):
  leave=int(ticks_ms()) + duration_ms
  for n in range(count):
    send_tms(0) # -> idle
  while int(ticks_ms())-leave < 0:
    send_tms(0) # -> idle
  send_tms(1) # -> select DR scan

# send SIR command (bytes)
# TAP should be in "select DR scan" state
# TAP returns to "select DR scan" state
@micropython.viper
def sir(buf):
  send_tms(1) # -> select IR scan
  send_tms(0) # -> capture IR
  send_tms(0) # -> shift IR
  send_read_buf_lsb1st(buf,1,0) # -> exit 1 IR
  send_tms0111() # -> select DR scan

# send SIR command (bytes)
# TAP should be in "select DR scan" state
# TAP returns to "select DR scan" state
# finish with n idle cycles during minimum of ms time
@micropython.viper
def sir_idle(buf, n:int, ms:int):
  send_tms(1) # -> select IR scan
  send_tms(0) # -> capture IR
  send_tms(0) # -> shift IR
  send_read_buf_lsb1st(buf,1,0) # -> exit 1 IR
  send_tms(0) # -> pause IR
  send_tms(1) # -> exit 2 IR
  send_tms(1) # -> update IR
  runtest_idle(n+1,ms) # -> select DR scan

@micropython.viper
def sdr(buf):
  send_tms(0) # -> capture DR
  send_tms(0) # -> shift DR
  send_read_buf_lsb1st(buf,1,0)
  send_tms0111() # -> select DR scan

@micropython.viper
def sdr_idle(buf, n:int, ms:int):
  send_tms(0) # -> capture DR
  send_tms(0) # -> shift DR
  send_read_buf_lsb1st(buf,1,0)
  send_tms(0) # -> pause DR
  send_tms(1) # -> exit 2 DR
  send_tms(1) # -> update DR
  runtest_idle(n+1, ms) # -> select DR scan

# sdr buffer will be overwritten with response
@micropython.viper
def sdr_response(buf):
  send_tms(0) # -> capture DR
  send_tms(0) # -> shift DR
  send_read_buf_lsb1st(buf,1,addressof(buf))
  send_tms0111() # -> select DR scan

def check_response(response, expected, mask=0xFFFFFFFF, message=""):
  if (response & mask) != expected:
    print("0x%08X & 0x%08X != 0x%08X %s" % (response,mask,expected,message))

def idcode():
  bitbang_jtag_on()
  led.on()
  reset_tap()
  runtest_idle(1,0)
  #sir(b"\xE0")
  id_bytes = bytearray(4)
  sdr_response(id_bytes)
  led.off()
  bitbang_jtag_off()
  return unpack("<I", id_bytes)[0]


# common JTAG open for both program and flash
def common_open():
  spi_jtag_on()
  hwspi.init(sck=Pin(jtagpin.tcknc)) # avoid TCK-glitch
  bitbang_jtag_on()
  led.on()
  reset_tap()
  runtest_idle(1,0)
  #sir(b"\xE0") # read IDCODE
  #sdr(pack("<I",0), expected=pack("<I",0), message="IDCODE")
  if discard:
    sir(b"\x1C") # LSC_PRELOAD: program Bscan register
    sdr(bytearray([0xFF for i in range(64)]))
    sir(b"\xC6") # ISC ENABLE: Enable SRAM programming mode
    sdr_idle(b"\x00",2,10)
    sir_idle(b"\x3C",2,1) # LSC_READ_STATUS
    status = bytearray(4)
    sdr_response(status)
    check_response(unpack("<I",status)[0], mask=0x24040, expected=0, message="FAIL status")
    sir(b"\x0E") # ISC_ERASE: Erase the SRAM
    sdr_idle(b"\x01",2,10)
  sir_idle(b"\x3C",2,1) # LSC_READ_STATUS
  status = bytearray(4)
  sdr_response(status)
  check_response(unpack("<I",status)[0], mask=0xB000, expected=0, message="FAIL status")

# call this before sending the flash image
# FPGA will enter flashing mode
# TAP should be in "select DR scan" state
@micropython.viper
def flash_open():
  common_open()
  reset_tap()
  runtest_idle(1,0)
  sir_idle(b"\xFF",32,0) # BYPASS
  sir(b"\x3A") # LSC_PROG_SPI
  sdr_idle(b"\xFE\x68",32,0)
  # ---------- flashing begin -----------
  # sdr(b"\x60") and other SPI FLASH commands
  # here are bitreverse() values of FLASH commands
  # found in datasheet. e.g.
  # \x1B -> 0xD8
  # \x60 -> 0x06 ...

@micropython.viper
def flash_wait_status(n:int):
  retry=n
  mask=1 # WIP bit (work-in-progress)
  send_tms(0) # -> capture DR
  send_tms(0) # -> shift DR
  swspi.write(read_status) # READ STATUS REGISTER
  swspi.readinto(status)
  while retry > 0:
    swspi.readinto(status)
    if (int(status[0]) & mask) == 0:
      break
    sleep_ms(1)
    retry -= 1
  send_tms(1) # -> exit 1 DR # exit at byte incomplete
  #send_int_msb1st(0,1,8) # exit at byte complete
  send_tms0111() # -> select DR scan
  if retry <= 0:
    print("error %d flash status 0x%02X & 0x%02X != 0" % (n,status[0],mask))

# call this after uploading all of the flash blocks,
# this will exit FPGA flashing mode and start the bitstream
@micropython.viper
def flash_close():
  # switch from SPI to bitbanging
  # ---------- flashing end -----------
  sdr(b"\x20") # SPI WRITE DISABLE
  sir_idle(b"\xFF",100,1) # BYPASS
  if discard:
    sir_idle(b"\x26",2,200) # ISC DISABLE
    sir_idle(b"\xFF",2,1) # BYPASS
    sir(b"\x79") # LSC_REFRESH reload the bitstream from flash
    sdr_idle(b"\x00\x00\x00",2,100)
  spi_jtag_off()
  reset_tap()
  led.off()
  bitbang_jtag_off()

def flash_send(seq):
  send_tms(0) # -> capture DR
  send_tms(0) # -> shift DR
  if len(seq)>1:
    swspi.write(seq[:-1]) # all except last byte
  send_int_msb1st(seq[-1],1,8) # last byte -> exit 1 DR
  send_tms0111() # -> select DR scan

def flash_sendrecv(send,recv):
  send_tms(0) # -> capture DR
  send_tms(0) # -> shift DR
  swspi.write(send)
  swspi.readinto(recv)
  #send_tms(1) # -> exit 1 DR # exit at byte incomplete
  send_int_msb1st(0,1,8) # complete dummy byte and exit
  send_tms0111() # -> select DR scan

def int2bin(a):
  bin=bytearray(8)
  for i in range(8):
    bin[7-i]=48+(a&1)
    a>>=1
  return bin

# write protection tool for IS25LP128
# https://www.issi.com/WW/pdf/IS25LP128.pdf
# prot=0: unprotect
# prot=6: protect first 2MB
def is25lp128_protect(prot=6):
  flash_open()
  # write function register
  # factory default is protecting the top (... - 0xFFFFFF)
  # now set to protect the bottom (0x000000 - ...).
  # NOTE: once set to protect the bottom, can't be reset back to protect the top
  # see datasheet p.17 t.6.5, p.50 s.8.19
  flash_send(b"\x06") # permanent write
  flash_wait_status(1011)
  flash_send(b"\x42\x02") # function reg = 0x02 TBS=1 OTP warning: once set, can't be reset!
  flash_wait_status(2012)
  # write status register value prot=6 to protect 2MB (0x000000 - 0x1FFFFF)
  # see datasheet p.14 t.6.3
  flash_send(b"\x06") # permanent write
  flash_wait_status(1021)
  flash_send(bytearray([1,prot<<2])) # status reg1=0x18 protect lower 2MB
  flash_wait_status(2022)
  flash_close()

def is25lp128_status():
  flash_open()
  status_reg1=bytearray(1)
  status_reg2=bytearray(1)
  flash_sendrecv(b"\x05",status_reg1)
  flash_sendrecv(b"\x48",status_reg2)
  flash_close()
  if discard:
    sleep_ms(1000)
  noyes_txt=("No","Yes")
  TBS=(status_reg2[0]>>1) & 1
  TBS_txt=("Top","Bottom")
  IRL=(status_reg2[0]>>4) & 15
  BP=(status_reg1[0]>>2) & 15
  QE=(status_reg1[0]>>6) & 1
  range_bytes = 32768<<BP
  if BP:
    if TBS:
      range="0x000000 - 0x%06X" % (range_bytes-1)
    else: # TBS=0
      range="0x%06X - 0xFFFFFF" % (0x1000000-range_bytes)
  else: # BP=0
    range="None"
  print("Read 0x05: Status Register = 0x%02X" % status_reg1[0])
  print(int2bin(status_reg1[0]).decode())
  print(".x...... QE  Quad Enable        : %s" % noyes_txt[QE])
  print("..xxxx.. BP  Protected Range    : %s" % range)
  print("Read 0x48: Function Register = 0x%02X" % status_reg2[0])
  print("%s OTP warning value 1 can't reset to 0" % int2bin(status_reg2[0]).decode())
  print("xxxx.... IRL Information Lock   : %d" % IRL)
  print("......x. TBS Top/Bottom Select  : %s" % TBS_txt[TBS])

# write protection tool for W25Q128JV, see datasheet p.18, p.26
# https://www.winbond.com/resource-files/w25q128jv%20revf%2003272018%20plus.pdf
# prot= 0: unprotect  
# prot=12: protecting first 2MB
def w25q128jv_protect(prot=12):
  flash_open()
  flash_wait_status(1021)
  flash_send(b"\x06") # permanent write
  flash_wait_status(1021)
  flash_send(bytearray([1,prot<<2])) # status reg1=0x30 protect lower 2MB
  flash_wait_status(1021)
  flash_send(b"\x06") # permanent write
  flash_wait_status(1021)
  flash_send(b"\x11\x60") # status reg3=0x60 WPS=0
  flash_wait_status(1021)
  #flash_send(b"\x50") # temporary write
  #flash_wait_status(1021)
  #flash_send(b"\x31\x01") # status reg2=0x01 lock status
  #flash_wait_status(1021)
  #status_reg1=bytearray(1)
  #flash_sendrecv(b"\x05",status_reg1)
  flash_close()
  #return status_reg1[0]

def w25q128jv_status():
  flash_open()
  status_reg1=bytearray(1)
  status_reg2=bytearray(1)
  status_reg3=bytearray(1)
  flash_sendrecv(b"\x05",status_reg1)
  flash_sendrecv(b"\x35",status_reg2)
  flash_sendrecv(b"\x15",status_reg3)
  flash_close()
  if discard:
    sleep_ms(1000)
  noyes_txt=("No","Yes")
  print("Read 0x05: Status Register-1 = 0x%02X" % status_reg1[0])
  SRP=(status_reg1[0]>>7) & 1
  SEC=(status_reg1[0]>>6) & 1
  SEC_txt=("64KB Blocks","4KB Sectors")
  TB=(status_reg1[0]>>5) & 1
  TB_txt=("Top","Bottom")
  BP=(status_reg1[0]>>2) & 7
  CMP=(status_reg2[0]>>6) & 1
  SEC_size=(128*1024,2048)
  range_bytes = SEC_size[SEC]<<BP
  if BP:
    if TB:
      if CMP:
        range="0x%06X - 0xFFFFFF" % (range_bytes)
      else: # CMP=0
        range="0x000000 - 0x%06X" % (range_bytes-1)
    else: # TB=0
      if CMP:
        range="0x000000 - 0x%06X" % (0xFFFFFF-range_bytes)
      else: # CMP=0
        range="0x%06X - 0xFFFFFF" % (0x1000000-range_bytes)
  else: # BP=0
    range="None"
  print(int2bin(status_reg1[0]).decode())
  print("x....... SRP Status Reg Protect by WP pin: %s" % noyes_txt[SRP])
  print(".x...... SEC Sector/Block Protect        : %s" % SEC_txt[SEC])
  print("..x..... TB  Top/Bottom Protect          : %s" % TB_txt[TB])
  print("...xxx.. BP  Protected Range             : %s" % range)
  print("Read 0x35: Status Register-2 = 0x%02X" % status_reg2[0])
  #CMP=(status_reg2[0]>>6) & 1
  LB=(status_reg2[0]>>3) & 7
  QE=(status_reg2[0]>>1) & 1
  SRL=status_reg2[0] & 1
  print("%s OTP warning value 1 can't reset to 0" % int2bin(status_reg2[0]).decode())
  print(".x...... CMP Complement Protect          : %s" % noyes_txt[CMP])
  print("..xxx... LB  Security Register Lock Bits : %d" % LB)
  print("......x. QE  Quad Enable                 : %s" % noyes_txt[QE])
  print(".......x SRL STATUS REGISTER LOCK        : %s" % noyes_txt[SRL])
  print("Read 0x15: Status Register-3 = 0x%02X" % status_reg3[0])
  print(int2bin(status_reg3[0]).decode())
  WPS=(status_reg3[0]>>2) & 1
  WPS_txt=("Defined by Register-1 and 2","Individual Sectors")
  DRV=(status_reg3[0]>>5) & 3
  DRV_strength=bytearray([100,75,50,25]) # %
  print(".xx..... DRV Output Driver Strength      : %d%%" % DRV_strength[DRV])
  print(".....x.. WPS Write Protection Scheme     : %s" % WPS_txt[WPS])

def detect():
  id=idcode()
  print("FPGA JTAG IDCODE 0x%08X" % id)
  if id==0:
    print("check pinout for v3.0.x/v3.1.x")
    return
  flash_open()
  manuf_dev_id=bytearray(2)
  flash_sendrecv(b"\x90\x00\x00\x00",manuf_dev_id)
  jedec_id=bytearray(3)
  flash_sendrecv(b"\x9F",jedec_id)
  unique_id=bytearray(8)
  flash_sendrecv(b"\x4B\x00\x00\x00\x00",unique_id)
  flash_close()
  if discard:
    sleep_ms(1000)
  print("Read 0x90: Manufacture/Device ID: %02X %02X" %
    (manuf_dev_id[0],manuf_dev_id[1]))
  print("Read 0x9F: JEDEC              ID: %02X %02X %02X" %
    (jedec_id[0],jedec_id[1],jedec_id[2]))
  print("Read 0x4B: Unique             ID: %02X %02X %02X %02X %02X %02X %02X %02X" %
    (unique_id[0],unique_id[1],unique_id[2],unique_id[3],unique_id[4],unique_id[5],unique_id[6],unique_id[7]))
  if manuf_dev_id==b"\x00\x00" or manuf_dev_id==b"\xFF\xFF":
    print("JTAG can't access SPI FLASH. Try")
    print("ESP32 passthru that drives WPn=1 HOLDn=1,")
    print("with SYSCONFIG MASTER_SPI_PORT=ENABLE in .lpf, without USRMCLK in .v")
    print("fujprog passthru.bit or edit this source: discard=1")
  if jedec_id==b"\xEF\x40\x18":
    print("Winbond W25Q128JV")
    if discard:
      sleep_ms(1000)
    w25q128jv_status()
    print("ecp5wp.w25q128jv_protect()")
    print("ecp5wp.w25q128jv_status()")
  if jedec_id==b"\x9D\x60\x18":
    print("ISSI IS25LP128")
    if discard:
      sleep_ms(1000)
    is25lp128_status()
    print("ecp5wp.is25lp128_protect()")
    print("ecp5wp.is25lp128_status()")
  if jedec_id==b"\x9D\x60\x16":
    print("ISSI IS25LP032")
    if discard:
      sleep_ms(1000)
    is25lp128_status()
    print("write protection not supported")

detect()



import sys
from pathlib import Path

__dir__ = Path(__file__).absolute().parent
# Remove current dir from sys.path, otherwise setuptools will peek up our
# module instead of system's.
sys.path.pop(0)
from setuptools import setup

sys.path.append(".")
import sdist_upip

setup(
    name='esp32ecp5',
    py_modules=['ecp5setup','ecp5','ecp5wp','sdraw','uftpd','wifiman'],
    version='1.0.9',
    description='MicroPython ESP32 JTAG programmer for ECP5 FPGA',
    long_description='Full featured ECP5 FPGA programmer with native support for ULX3S boards',
    long_description_content_type="text/markdown",
    keywords='JTAG, FPGA, ECP5, WiFi, FTP, FLASH, SD',
    url='https://github.com/emard/esp32ecp5',
    author='EMARD',
    author_email='vordah@gmail.com',
    maintainer='EMARD',
    maintainer_email='vordah@gmail.com',
    license='MIT',
    cmdclass={'sdist': sdist_upip.sdist},
    project_urls={
        'Bug Reports': 'https://github.com/emard/esp32ecp5/issues',
        'Documentation': 'https://github.com/emard/esp32ecp5/blob/master/README.md',
        'Source': 'https://github.com/emard/esp32ecp5',
    },
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: Implementation :: MicroPython',
        'License :: OSI Approved :: MIT License',
    ],
)



# micropython ESP32
# ECP5 JTAG programmer

# AUTHOR=EMARD
# LICENSE=BSD

from time import ticks_ms, sleep_ms
from machine import SPI, SoftSPI, Pin, freq
from micropython import const
from struct import unpack
from uctypes import addressof
import jtagpin

freq(240000000) # Hz CPU clk frequency
spi_freq = const(20000000) # Hz JTAG clk frequency
flash_read_size = const(2048)
flash_write_size = const(256)
flash_erase_size = const(4096)
flash_erase_cmd = { 4096:0x20, 32768:0x52, 65536:0xD8, 262144:0xD8 } # erase commands from FLASH PDF
flash_era = bytearray([flash_erase_cmd[flash_erase_size],0,0])
flash_req=bytearray(4)
read_status=bytearray([5])
status=bytearray(1)
#rb=bytearray(256) # reverse bits
#init_reverse_bits()

def bitbang_jtag_on():
  global tck,tms,tdi,tdo,led
  led=Pin(jtagpin.led,Pin.OUT)
  tms=Pin(jtagpin.tms,Pin.OUT)
  tck=Pin(jtagpin.tck,Pin.OUT)
  tdi=Pin(jtagpin.tdi,Pin.OUT)
  tdo=Pin(jtagpin.tdo,Pin.IN)

def bitbang_jtag_off():
  global tck,tms,tdi,tdo,led
  led=Pin(jtagpin.led,Pin.IN)
  tms=Pin(jtagpin.tms,Pin.IN)
  tck=Pin(jtagpin.tck,Pin.IN)
  tdi=Pin(jtagpin.tdi,Pin.IN)
  tdo=Pin(jtagpin.tdo,Pin.IN)
  a = led.value()
  a = tms.value()
  a = tck.value()
  a = tdo.value()
  a = tdi.value()
  del led
  del tms
  del tck
  del tdi
  del tdo

# initialize both hardware accelerated SPI
# software SPI on the same pins
def spi_jtag_on():
  global hwspi,swspi
  hwspi=SPI(2, baudrate=spi_freq, polarity=1, phase=1, bits=8, firstbit=SPI.MSB, sck=Pin(jtagpin.tck), mosi=Pin(jtagpin.tdi), miso=Pin(jtagpin.tdo))
  swspi=SoftSPI(baudrate=spi_freq, polarity=1, phase=1, bits=8, firstbit=SPI.MSB, sck=Pin(jtagpin.tck), mosi=Pin(jtagpin.tdi), miso=Pin(jtagpin.tdo))

def spi_jtag_off():
  global hwspi,swspi
  hwspi.deinit()
  del hwspi
  swspi.deinit()
  del swspi

# print bytes reverse - appears the same as in SVF file
#def print_hex_reverse(block, head="", tail="\n"):
#  print(head, end="")
#  for n in range(len(block)):
#    print("%02X" % block[len(block)-n-1], end="")
#  print(tail, end="")

#@micropython.viper
#def init_reverse_bits():
#  #p8rb=ptr8(addressof(rb))
#  p8rb=memoryview(rb)
#  for i in range(256):
#    v=i
#    r=0
#    for j in range(8):
#      r<<=1
#      r|=v&1
#      v>>=1
#    p8rb[i]=r

@micropython.viper
def send_tms(val:int,n:int):
  if val:
    tms.on()
  else:
    tms.off()
  for i in range(n):
    tck.off()
    tck.on()

# exit 1 DR -> select DR scan
def send_tms0111():
  send_tms(0,1) # -> pause DR
  send_tms(1,3) # -> exit 2 DR -> update DR -> select DR scan

@micropython.viper
def send_read_buf_lsb1st(buf, last:int, w:ptr8):
  p = ptr8(addressof(buf))
  l = int(len(buf))
  val = 0
  tms.off()
  for i in range(l-1):
    byte = 0
    val = p[i]
    for nf in range(8):
      if (val >> nf) & 1:
        tdi.on()
      else:
        tdi.off()
      tck.off()
      tck.on()
      if tdo.value():
        byte |= 1 << nf
    if int(w):
      w[i] = byte # write byte
  byte = 0
  val = p[l-1] # read last byte
  for nf in range(7): # first 7 bits
    if (val >> nf) & 1:
      tdi.on()
    else:
      tdi.off()
    tck.off()
    tck.on()
    if tdo.value():
      byte |= 1 << nf
  # last bit
  if last:
    tms.on()
  if (val >> 7) & 1:
    tdi.on()
  else:
    tdi.off()
  tck.off()
  tck.on()
  if tdo.value():
    byte |= 1 << 7
  if int(w):
    w[l-1] = byte # write last byte

@micropython.viper
def send_int_msb1st(val:int, last:int, bits:int):
  tms.off()
  for nf in range(bits-1):
    if (val >> (7-nf)) & 1:
      tdi.on()
    else:
      tdi.off()
    tck.off()
    tck.on()
  if last:
    tms.on()
  if val & 1:
    tdi.on()
  else:
    tdi.off()
  tck.off()
  tck.on()

# TAP should be in "idle" state
# TAP returns to "select DR scan" state
@micropython.viper
def runtest_idle(count:int, duration_ms:int):
  leave=int(ticks_ms()) + duration_ms
  send_tms(0,count) # -> idle
  while int(ticks_ms())-leave < 0:
    send_tms(0,1) # -> idle
  send_tms(1,1) # -> select DR scan

# send SIR command (bytes)
# TAP should be in "select DR scan" state
# TAP returns to "select DR scan" state
@micropython.viper
def sir(buf):
  send_tms(1,1) # -> select IR scan
  send_tms(0,2) # -> capture IR -> shift IR
  send_read_buf_lsb1st(buf,1,0) # -> exit 1 IR
  send_tms0111() # -> select DR scan

# send SIR command (bytes)
# TAP should be in "select DR scan" state
# TAP returns to "select DR scan" state
# finish with n idle cycles during minimum of ms time
@micropython.viper
def sir_idle(buf, n:int, ms:int):
  send_tms(1,1) # -> select IR scan
  send_tms(0,2) # -> capture IR -> shift IR
  send_read_buf_lsb1st(buf,1,0) # -> exit 1 IR
  send_tms(0,1) # -> pause IR
  send_tms(1,2) # -> exit 2 IR -> update IR
  runtest_idle(n+1,ms) # -> select DR scan

@micropython.viper
def sdr(buf):
  send_tms(0,2) # -> capture DR -> shift DR
  send_read_buf_lsb1st(buf,1,0)
  send_tms0111() # -> select DR scan

@micropython.viper
def sdr_idle(buf, n:int, ms:int):
  send_tms(0,2) # -> capture DR -> shift DR
  send_read_buf_lsb1st(buf,1,0)
  send_tms(0,1) # -> pause IR
  send_tms(1,2) # -> exit 2 IR -> update IR
  runtest_idle(n+1, ms) # -> select DR scan

# sdr buffer will be overwritten with response
@micropython.viper
def sdr_response(buf):
  send_tms(0,2) # -> capture DR -> shift DR
  send_read_buf_lsb1st(buf,1,addressof(buf))
  send_tms0111() # -> select DR scan

def check_response(response, expected, mask=0xFFFFFFFF, message=""):
  if (response & mask) != expected:
    print("0x%08X & 0x%08X != 0x%08X %s" % (response,mask,expected,message))

def idcode():
  bitbang_jtag_on()
  led.on()
  send_tms(1,6) # -> Test Logic Reset
  runtest_idle(1,0)
  #sir(b"\xE0")
  id_bytes = bytearray(4)
  sdr_response(id_bytes)
  led.off()
  bitbang_jtag_off()
  return unpack("<I", id_bytes)[0]

# common JTAG open for both program and flash
def common_open():
  spi_jtag_on()
  hwspi.init(sck=Pin(jtagpin.tcknc)) # avoid TCK-glitch
  bitbang_jtag_on()
  led.on()
  send_tms(1,6) # -> Test Logic Reset
  runtest_idle(1,0)
  #sir(b"\xE0") # read IDCODE
  #sdr(pack("<I",0), expected=pack("<I",0), message="IDCODE")
  sir(b"\x1C") # LSC_PRELOAD: program Bscan register
  sdr(bytearray([0xFF for i in range(64)]))
  sir(b"\xC6") # ISC ENABLE: Enable SRAM programming mode
  sdr_idle(b"\x00",2,10)
  sir_idle(b"\x3C",2,1) # LSC_READ_STATUS
  status = bytearray(4)
  sdr_response(status)
  check_response(unpack("<I",status)[0], mask=0x24040, expected=0, message="FAIL status")
  sir(b"\x0E") # ISC_ERASE: Erase the SRAM
  sdr_idle(b"\x01",2,10)
  sir_idle(b"\x3C",2,1) # LSC_READ_STATUS
  status = bytearray(4)
  sdr_response(status)
  check_response(unpack("<I",status)[0], mask=0xB000, expected=0, message="FAIL status")

# call this before sending the bitstram
# FPGA will enter programming mode
# after this TAP will be in "shift DR" state
def prog_open():
  common_open()
  sir(b"\x46") # LSC_INIT_ADDRESS
  sdr_idle(b"\x01",2,10)
  sir(b"\x7A") # LSC_BITSTREAM_BURST
  # ---------- bitstream begin -----------
  # manually walk the TAP
  # we will be sending one long DR command
  send_tms(0,2) # -> capture DR -> shift DR
  # switch from bitbanging to SPI mode
  hwspi.init(sck=Pin(jtagpin.tck)) # 1 TCK-glitch? TDI=0
  # we are lucky that format of the bitstream tolerates
  # any leading and trailing junk bits. If it weren't so,
  # HW SPI JTAG acceleration wouldn't work.
  # to upload the bitstream:
  # FAST SPI mode
  #hwspi.write(block)
  # SLOW bitbanging mode
  #for byte in block:
  #  send_int_msb1st(byte,0)

def prog_stream_done():
  # switch from hardware SPI to bitbanging done after prog_stream()
  hwspi.init(sck=Pin(jtagpin.tcknc)) # avoid TCK-glitch
  spi_jtag_off()

# call this after uploading all of the bitstream blocks,
# this will exit FPGA programming mode and start the bitstream
# returns status True-OK False-Fail
def prog_close():
  bitbang_jtag_on()
  send_tms(1,1) # -> exit 1 DR
  send_tms(0,1) # -> pause DR
  send_tms(1,2) # -> exit 2 DR -> update DR
  #send_tms(0,1) # -> idle, disabled here as runtest_idle does the same
  runtest_idle(100,10)
  # ---------- bitstream end -----------
  sir_idle(b"\xC0",2,1) # read usercode
  usercode = bytearray(4)
  sdr_response(usercode)
  check_response(unpack("<I",usercode)[0],expected=0,message="FAIL usercode")
  sir_idle(b"\x26",2,200) # ISC DISABLE
  sir_idle(b"\xFF",2,1) # BYPASS
  sir(b"\x3C") # LSC_READ_STATUS
  status = bytearray(4)
  sdr_response(status)
  status = unpack("<I",status)[0]
  check_response(status,mask=0x2100,expected=0x100,message="FAIL status")
  done = True
  if (status & 0x2100) != 0x100:
    done = False
  send_tms(1,6) # -> Test Logic Reset
  led.off()
  bitbang_jtag_off()
  return done

# call this before sending the flash image
# FPGA will enter flashing mode
# TAP should be in "select DR scan" state
@micropython.viper
def flash_open():
  common_open()
  send_tms(1,6) # -> Test Logic Reset
  runtest_idle(1,0)
  sir_idle(b"\xFF",32,0) # BYPASS
  sir(b"\x3A") # LSC_PROG_SPI
  sdr_idle(b"\xFE\x68",32,0)
  # ---------- flashing begin -----------
  # sdr("\x60") and other SPI FLASH commands
  # here are bitreverse() values of FLASH commands
  # found in datasheet. e.g.
  # \x1B -> 0xD8
  # \x60 -> 0x06 ...

@micropython.viper
def flash_wait_status(n:int):
  retry=n
  mask=1 # WIP bit (work-in-progress)
  send_tms(0,2) # -> capture DR -> shift DR
  swspi.write(read_status) # READ STATUS REGISTER
  swspi.readinto(status)
  while retry > 0:
    swspi.readinto(status)
    if (int(status[0]) & mask) == 0:
      break
    sleep_ms(1)
    retry -= 1
  send_tms(1,1) # -> exit 1 DR # exit at byte incomplete
  #send_int_msb1st(0,1,8) # exit at byte complete
  send_tms0111() # -> select DR scan
  if retry <= 0:
    print("error %d flash status 0x%02X & 0x%02X != 0" % (n,status[0],mask))

@micropython.viper
def flash_erase_block(addr:int):
  sdr(b"\x60") # SPI WRITE ENABLE
  flash_wait_status(1001)
  p8=ptr8(addressof(flash_era))
  p8[1]=addr>>16
  p8[2]=addr>>8
  send_tms(0,2) # -> capture DR -> shift DR
  swspi.write(flash_era) # except LSB
  send_int_msb1st(addr,1,8) # last LSB byte -> exit 1 DR
  send_tms0111() # -> select DR scan
  flash_wait_status(2002)

@micropython.viper
def flash_write_block(block, last:int, addr:int):
  sdr(b"\x60") # SPI WRITE ENABLE
  flash_wait_status(1003)
  p8=ptr8(addressof(flash_req))
  p8[0]=2
  p8[1]=addr>>16
  p8[2]=addr>>8
  p8[3]=addr
  send_tms(0,2) # -> capture DR -> shift DR
  swspi.write(flash_req)
  swspi.write(block) # whole block
  send_int_msb1st(last,1,8) # last byte -> exit 1 DR
  send_tms0111() # -> select DR scan
  flash_wait_status(1004)

# data is bytearray of to-be-read length
@micropython.viper
def flash_read_block(data, addr:int):
  p8=ptr8(addressof(flash_req))
  p8[0]=3
  p8[1]=addr>>16
  p8[2]=addr>>8
  p8[3]=addr
  send_tms(0,2) # -> capture DR -> shift DR
  swspi.write(flash_req) # send SPI FLASH read command and address and dummy byte
  swspi.readinto(data) # retrieve whole block
  send_int_msb1st(0,1,8) # dummy read byte -> exit 1 DR
  send_tms0111() # -> select DR scan

# call this after uploading all of the flash blocks,
# this will exit FPGA flashing mode and start the bitstream
@micropython.viper
def flash_close():
  # switch from SPI to bitbanging
  # ---------- flashing end -----------
  sdr(b"\x20") # SPI WRITE DISABLE
  sir_idle(b"\xFF",100,1) # BYPASS
  sir_idle(b"\x26",2,200) # ISC DISABLE
  sir_idle(b"\xFF",2,1) # BYPASS
  sir(b"\x79") # LSC_REFRESH reload the bitstream from flash
  sdr_idle(b"\x00\x00\x00",2,100)
  spi_jtag_off()
  send_tms(1,6) # -> Test Logic Reset
  led.off()
  bitbang_jtag_off()

def stopwatch_start():
  global stopwatch_ms
  stopwatch_ms = ticks_ms()

def stopwatch_stop(bytes_uploaded):
  global stopwatch_ms
  elapsed_ms = ticks_ms() - stopwatch_ms
  transfer_rate_MBps = 0
  if elapsed_ms > 0:
    transfer_rate_kBps = bytes_uploaded // elapsed_ms
  print("%d bytes uploaded in %d ms (%d kB/s)" % (bytes_uploaded, elapsed_ms, transfer_rate_kBps))

def prog_stream(filedata, blocksize=16384):
  prog_open()
  bytes_uploaded = 0
  stopwatch_start()
  block = bytearray(blocksize)
  while True:
    if filedata.readinto(block):
      hwspi.write(block)
      bytes_uploaded += len(block)
    else:
      break
  stopwatch_stop(bytes_uploaded)
  prog_stream_done()

def open_file(filename, gz=False):
  filedata = open(filename, "rb")
  if gz:
    import uzlib
    return uzlib.DecompIO(filedata,31)
  return filedata

def open_web(url, gz=False):
  import socket
  _, _, host, path = url.split('/', 3)
  port = 80
  if ( len(host.split(':')) == 2 ):
    host, port = host.split(':', 2)
    port = int(port)
  print("host = %s, port = %d, path = %s" % (host, port, path))
  addr = socket.getaddrinfo(host, port)[0][-1]
  s = socket.socket()
  s.connect(addr)
  s.send(bytes('GET /%s HTTP/1.0\r\nHost: %s\r\nAccept:  image/*\r\n\r\n' % (path, host), 'utf8'))
  for i in range(100): # read first 100 lines searching for
    if len(s.readline()) < 3: # first empty line (contains "\r\n")
      break
  if gz:
    import uzlib
    return uzlib.DecompIO(s,31)
  return s

# data is bytearray of to-be-read length
def flash_read(data, addr=0):
  flash_open()
  flash_read_block(data, addr)
  flash_close()

# accelerated compare flash and file block
# return value
# 0-must nothing, 1-must erase, 2-must write, 3-must erase and write
@micropython.viper
def compare_flash_file_buf(flash_b, file_b, must:int)->int:
  flash_block = ptr8(addressof(flash_b))
  file_block = ptr8(addressof(file_b))
  l = int(len(file_b))
  for i in range(l):
    if (flash_block[i] & file_block[i]) != file_block[i]:
      must = 1
  if must: # erase will reset all bytes to 0xFF
    for i in range(l):
      if file_block[i] != 0xFF:
        must = 3
  else: # no erase
    for i in range(l):
      if flash_block[i] != file_block[i]:
        must = 2
  return must

# clever = read-compare-erase-write
# prevents flash wear when overwriting the same data
# 4K erase block is max that fits on ESP32-WROOM
# returns status True-OK False-Fail
def flash_stream(filedata, addr=0):
  flash_open()
  addr_mask = flash_erase_size-1
  if addr & addr_mask:
    print("addr must be rounded to flash_erase_size = %d bytes (& 0x%06X)" % (flash_erase_size, 0xFFFFFF & ~addr_mask))
    return False
  addr = addr & 0xFFFFFF & ~addr_mask # rounded to even 64K (erase block)
  bytes_uploaded = 0
  stopwatch_start()
  #if 1:
  #  print("erase whole FLASH (max 90s)")
  #  sdr(b"\x60") # SPI WRITE ENABLE
  #  flash_wait_status(1005)
  #  sdr(b"\xE3") # BULK ERASE (whole chip) rb[0x60]=0x06 or rb[0xC7]=0xE3
  #  flash_wait_status(90000)
  count_total = 0
  count_erase = 0
  count_write = 0
  file_block = bytearray(flash_erase_size)
  flash_block = bytearray(flash_read_size)
  file_blockmv=memoryview(file_block)
  progress_char="."
  while filedata.readinto(file_block):
    led.value((bytes_uploaded >> 12)&1)
    retry = 3
    while retry > 0:
      must = 0
      flash_rd = 0
      while flash_rd<flash_erase_size:
        flash_read_block(flash_block,addr+bytes_uploaded+flash_rd)
        must = compare_flash_file_buf(flash_block,file_blockmv[flash_rd:flash_rd+flash_read_size],must)
        flash_rd+=flash_read_size
      write_addr = addr+bytes_uploaded
      if must == 0:
        if (write_addr & 0xFFFF) == 0:
          print("\r0x%06X %dK %c" % (write_addr, flash_erase_size>>10, progress_char),end="")
        else:
          print(progress_char,end="")
        progress_char="."
        count_total += 1
        bytes_uploaded += len(file_block)
        break
      retry -= 1
      if must & 1: # must_erase:
        #print("from 0x%06X erase %dK" % (write_addr, flash_erase_size>>10),end="\r")
        flash_erase_block(write_addr)
        count_erase += 1
        progress_char = "e"
      if must & 2: # must_write:
        #print("from 0x%06X write %dK" % (write_addr, flash_erase_size>>10),end="\r")
        block_addr = 0
        next_block_addr = 0
        while next_block_addr < len(file_block):
          next_block_addr = block_addr+flash_write_size
          flash_write_block(file_blockmv[block_addr:next_block_addr-1], file_blockmv[next_block_addr-1], write_addr)
          write_addr += flash_write_size
          block_addr = next_block_addr
        count_write += 1
        progress_char = "w"
      #if not verify:
      #  count_total += 1
      #  bytes_uploaded += len(file_block)
      #  break
    if retry <= 0:
      break
  print("\r",end="")
  stopwatch_stop(bytes_uploaded)
  print("%dK blocks: %d total, %d erased, %d written." % (flash_erase_size>>10, count_total, count_erase, count_write))
  return retry > 0 # True if successful

def filedata_gz(filepath):
  gz = filepath.endswith(".gz")
  if filepath.startswith("http://") or filepath.startswith("/http:/"):
    filedata = open_web(filepath, gz)
  else:
    filedata = open_file(filepath, gz)
  return filedata, gz

def prog(filepath, close=True):
  filedata, gz = filedata_gz(filepath)
  if filedata:
    if gz:
      prog_stream(filedata,blocksize=4096)
    else:
      prog_stream(filedata,blocksize=16384)
    # NOTE now the SD card can be released before bitstream starts
    if close:
      return prog_close() # start the bitstream
    return True
  return False

def flash(filepath, addr=0, close=True):
  filedata, gz = filedata_gz(filepath)
  if filedata:
    status=flash_stream(filedata,addr)
    # NOTE now the SD card can be released before bitstream starts
    if close:
      flash_close() # start the bitstream
    return status
  return False

def flashrd(addr=0, length=1):
  data = bytearray(length)
  flash_read(data, addr)
  return data

def passthru():
  id = idcode()
  if id != 0 and id != 0xFFFFFFFF:
    filepath = "passthru%08x.bit.gz" % id
    print("ecp5.prog(\"%s\")" % filepath)
    filedata = open_file(filepath, gz=True)
    prog_stream(filedata,blocksize=4096)
    return prog_close()
  return False

def help():
  print("usage:")
  print("ecp5.flash(\"blink.bit.gz\", addr=0x000000)")
  print("ecp5.flashrd(addr=0x000000, length=1)")
  print("ecp5.prog(\"http://192.168.4.2/blink.bit\")")
  print("ecp5.prog(\"blink.bit.gz\") # gzip -9 blink.bit")
  print("ecp5.passthru()")
  print("\"0x%08X\" % ecp5.idcode()")
  print("0x%08X" % idcode())
